@@ -7,7 +7,7 @@
 package DiaColloDB::Relation::TDF;
 use DiaColloDB::Relation;
 use DiaColloDB::Relation::TDF::Query;
-use DiaColloDB::Utils qw(:pack :fcntl :file :math :json :list :pdl :temp :env :run);
+use DiaColloDB::Utils qw(:pack :fcntl :file :math :json :list :pdl :temp :env :run :jobs);
 use DiaColloDB::PackedFile;
 #use DiaColloDB::Temp::Hash;
 #use DiaColloDB::Temp::Array;
@@ -30,8 +30,12 @@ our @ISA = qw(DiaColloDB::Relation);
 BEGIN {
   no warnings 'once';
   $PDL::BIGPDL = 1; ##-- avoid 'Probably false alloc of over 1Gb PDL' errors
+
+  PDL::no_clone_skip_warning() ##-- silence 'check out PDL::Parallel::threads' warning
+      if (UNIVERSAL::can('PDL','no_clone_skip_warning'));
 }
 
+
 ##==============================================================================
 ## Constructors etc.
 
@@ -205,9 +209,9 @@ sub open {
   my ($hdr); ##-- save header, for version-checking
   if (fcread($flags) && !fctrunc($flags)) {
     $hdr = $vs->readHeader()
-      or $vs->logconess("failed to read header data from '$vs->{base}.hdr': $!");
+      or $vs->logconfess("failed to read header data from '$vs->{base}.hdr': $!");
     $vs->loadHeaderData($hdr)
-      or $vs->logconess("failed to load header data from '$vs->{base}.hdr': $!");
+      or $vs->logconfess("failed to load header data from '$vs->{base}.hdr': $!");
   }
 
   ##-- check compatibility
@@ -381,22 +385,26 @@ sub create {
   my $len_date  = packsize($pack_date,0);
   my %tmpargs   = (UNLINK=>!$coldb->{keeptmp});
   my $tdm0file  = $wdmfile || "$vsdir/tdm0.dat";   # txt ~ "$ai0 $ai1 ... $aiN $doci $f"
-  my $tdm0fh    = $wdmfile ? undef : (opencmd("|-:raw", 'sort', (map {"-nk$_"} (1..($NA+1))), "-o", $tdm0file)
-				      or $vs->logconfess("$logas: failed to create pipe to sort for tempfile $tdm0file: $!"));
+  my ($tdm0fh);
+  if (!$wdmfile) {
+    ##-- v0.12.012_03 : use temporary $tdm0file.tmp so that later (sort TMPFILE) can do its parallel thing
+    CORE::open($tdm0fh, ">:raw", "$tdm0file.tmp")
+        or $vs->logconfess("$logas: failed to open temporary file $tdm0file.tmp: $!");
+  }
 
   ##-- create cat-wise piddle files c2date.pdl, c2d.pdl
-  my $c2datefile = "$vsdir/c2date.pdl";					##-- c2date ($NC): [$ci]   -> $date
+  my $c2datefile = "$vsdir/c2date.pdl";				##-- c2date ($NC): [$ci]   -> $date
   CORE::open(my $c2datefh, ">:raw", $c2datefile)
     or $vs->logconfess("$logas: failed to create piddle file $c2datefile: $!");
   writePdlHeader("$c2datefile.hdr", ushort, 1, $NC)
     or $vs->logconfess("$logas: failed to write piddle header $c2datefile.hdr: $!");
-  my $c2dfile = "$vsdir/c2d.pdl";						##-- c2d  (2,$NC): [0,$ci] => $di_off, [1,$ci] => $di_len
+  my $c2dfile = "$vsdir/c2d.pdl";				##-- c2d  (2,$NC): [0,$ci] => $di_off, [1,$ci] => $di_len
   CORE::open(my $c2dfh, ">:raw", $c2dfile)
       or $vs->logconfess("$logas: failed to create piddle file $c2dfile: $!");
   writePdlHeader("$c2dfile.hdr", $itype, 2, 2,$NC)
     or $vs->logconfess("$logas: failed to write piddle header $c2dfile.hdr: $!");
 
-  ##-- create: simulate DocClassify::Mapper::trainCorpus(): populate tdm0.*
+  ##-- create: tdf-sig: simulate DocClassify::Mapper::trainCorpus(): populate tdm0.*, c2date.*, c2d.*
   $vs->vlog($logCreate, "$logas: processing input documents [NA=$NA, NC=$nfiles]");
   my $json   = DiaColloDB::Utils->jsonxs();
   my $minDocSize = $vs->{minDocSize} = max2(($vs->{minDocSize}//0),1);
@@ -436,7 +444,7 @@ sub create {
       $mdata->{vals}[$docid] = $mvali;
     }
 
-    ##-- parse document signatures into $tdm0file
+    ##-- parse document signatures into $tdm0file.tmp (unsorted for now)
     #$vs->debug("sigs: id=$docid/$NC ; doc=$doclabel");
     if (defined $vtokfh) {
       $sigj_in = $sigi_in + $doc->{nsigs};
@@ -473,10 +481,18 @@ sub create {
   ##-- cleanup
   $c2dfh->close() or $vs->logconfess("$logas: close failed for tempfile $c2dfile: $!");
   $c2datefh->close() or $vs->logconfess("$logas: close failed for tempfile $c2datefile: $!");
-  !$tdm0fh or $tdm0fh->close() or $vs->logconfess("$logas: close failed for tempfile $tdm0file: $!");
+  !$tdm0fh or $tdm0fh->close() or $vs->logconfess("$logas: close failed for tempfile $tdm0file.tmp: $!");
   tied(@{$_->{vals}})->flush() foreach (values %meta);
 
-  ##-- create: filter: by term-frequency (default: use coldb term-filtering only)
+  ##-- v0.12.012_03: sort tdm0file.tmp -> tdm0file
+  if (!$wdmfile) {
+    runcmd('sort', (map {"-nk$_"} (1..($NA+1))), sortJobs(), "$tdm0file.tmp", "-o", $tdm0file)==0
+      or $vs->logconfess("$logas: failed to sort for $tdm0file.tmp: $!");
+    CORE::unlink("$tdm0file.tmp")
+      or $vs->logconfess("$logas: failed to unlink $tdm0file.tmp: $!");
+  }
+
+  ##-- create: filter: filter by term-frequency (default: use coldb term-filtering only)
   $vs->{minFreq} //= 0;
   my ($wbad);
   if ($vs->{minFreq} > 0) {
@@ -516,11 +532,10 @@ sub create {
     $vs->vlog($logCreate, "$logas: filter: will prune $nwbad of $NT0 term tuple type(s) ($pwbad)");
   }
 
-  ##-- create: filter: by doc-frequency
+  ##-- create: tdf-filter: filter by doc-frequency
   $vs->{minDocFreq} //= 0;
   if ($vs->{minDocFreq} > 0) {
     $vs->vlog($logCreate, "$logas: filter: by doc-frequency (minDocFreq=$vs->{minDocFreq})");
-    #env_push(LC_ALL=>'C');
     my $cmdfh = opencmd("cut -d\" \" -f-$NA $tdm0file | uniq -c |")
       or $vs->logconfess("$logas: failed to open pipe from sort for doc-frequency filter");
     $wbad //= tmphash("$vsdir/wbad", %tmpargs);
@@ -538,7 +553,6 @@ sub create {
 	++$NT1;
       }
     }
-    #env_pop();
     CORE::close($cmdfh);
 
     my $nwbad = ($NT0-$NT1);
@@ -546,7 +560,7 @@ sub create {
     $vs->vlog($logCreate, "$logas: filter: will prune $nwbad of $NT0 term tuple type(s) ($pwbad)");
   }
 
-  ##-- create: term-enum $tvals
+  ##-- create: filter: term-enum $tvals (+temporary %$ts2i)
   $vs->vlog($logCreate, "$logas: extracting term tuples");
   my $NT   = 0;
   my $NT0  = 0;
@@ -555,16 +569,22 @@ sub create {
   my $tvalsfile = "$vsdir/tvals.pdl";
   CORE::open(my $tvalsfh, ">:raw", $tvalsfile)
     or $vs->logconfess("$logas: open failed for term-values piddle $tvalsfile: $!");
-  my $ts2i = tmphash("$vsdir/ts2i", %tmpargs);
 
-  ##-- create: always include "null" term
+  ##-- %$ts2i: text proto-enum: "$ai1 $ai2 ... $aiN" => $ti
+  #$vs->vlog("debug", "$logas: using temporary term translation table $vsdir/ts2i.*");
+  #my $ts2i = tmphash("$vsdir/ts2i", %tmpargs);
+  ##--
+  $vs->vlog("debug", "$logas: using in-memory term translation hash");
+  my $ts2i = {};
+
+  ##-- create: filter: term-enum: always include "null" term
   {
     my @tnull = map {0} (1..$NA);
     $ts2i->{join(' ', @tnull)} = 0;
     $tvalsfh->print(pack($pack_ix, @tnull));
   }
 
-  ##-- create: enumerate "normal" terms in $tvalsfile
+  ##-- create: filter: term-enum: enumerate "normal" terms in $tvalsfile
   while (defined($_=<$ttxtfh>)) {
     chomp;
     ++$NT0;
@@ -585,7 +605,7 @@ sub create {
   my $pprunet = $NT0 ? sprintf("%.2f%%", 100*($NT0-$NT)/$NT0) : 'nan%';
   $vs->vlog($logCreate, "$logas: extracted $NT of $NT0 unique term tuples ($pprunet pruned)");
 
-  ##-- create: tdm0: ccs
+  ##-- create: tdf-matrix: tdm0: ccs
   my $ND  = $sigi_out;
   $vs->vlog($logCreate, "$logas: creating raw term-document matrix $vsdir/tdm.* (NT=$NT, ND=$ND)");
   my $ixfile = "$vsdir/tdm.ix";
@@ -614,11 +634,12 @@ sub create {
   CORE::close($ixfh)
       or $vs->logconfess("$logas: close failed for tdm index file $ixfile: $!");
   CORE::close($tdm0fh);
+  undef $ts2i;
   my $density  = sprintf("%.2g%%", $nnz / ($ND*$NT));
   my $pprunenz = $nnz0 ? sprintf("%.2f%%", 100*($nnz0-$nnz)/$nnz0) : 'nan%';
   $vs->vlog($logCreate, "$logas: created raw term-document matrix (density=$density, $pprunenz pruned)");
 
-  ##-- tdm0: read in as piddle
+  ##-- create: tdm0: read in as piddle
   writePdlHeader("$vsdir/tdm.ix.hdr", $itype, 2, 2,$nnz)
     or $vs->logconfess("$logas: failed to save tdm index header $vsdir/tdm.ix.hdr: $!");
   writePdlHeader("$vsdir/tdm.nz.hdr", $vtype, 1, $nnz+1)
@@ -628,7 +649,7 @@ sub create {
   defined(my $tdm = readPdlFile("$vsdir/tdm", class=>'PDL::CCS::Nd'))
     or $vs->logconfess("$logas: failed to map CCS term-document matrix from $vsdir/tdm.*");
 
-  ##-- create: N
+  ##-- create: tdm0: N
   $vs->{N} = $tdm->_vals->sum;
   $vs->vlog($logCreate, "$logas: computed total corpus size = $vs->{N}");
 
@@ -639,12 +660,12 @@ sub create {
   $c2d->slice("(1),")->rld(sequence($itype,$NC), my $d2c=mmzeroes("$vsdir/d2c.pdl",$itype,$ND));
   undef $c2d;
 
-  ##-- create: cf: ($NC): [$ci] -> f($ci)
+  ##-- create: aux: cf: ($NC): [$ci] -> f($ci)
   $vs->vlog($logCreate, "$logas: creating cat-frequency piddle $vsdir/cf.pdl (NC=$NC)");
   $tdm->_nzvals->indadd( $d2c->index($tdm->_whichND->slice("(1),")), my $cf=mmzeroes("$vsdir/cf.pdl",$vtype,$NC));
   #undef $cf;
 
-  ##-- create: yf: ($NY): [$yi] -> f($yi)
+  ##-- create: aux: yf: ($NY): [$yi] -> f($yi)
   defined(my $c2date = readPdlFile("$c2datefile"))
     or $vs->logconfess("$logas: failed to mmap $c2datefile");
   my ($ymin,$ymax) = $c2date->minmax;
@@ -657,7 +678,7 @@ sub create {
   undef $yf;
   undef $cf;
 
-  ##-- create: tym: ($NT,$NY): [$ti,$yi] -> f($ti,$yi)
+  ##-- create: aux: tym: ($NT,$NY): [$ti,$yi] -> f($ti,$yi)
   $vs->vlog($logCreate, "$logas: creating term-year matrix $vsdir/tym.*");
 
   ##-- tym: create using local memory-optimized pdl-pp method
@@ -670,12 +691,12 @@ sub create {
   writeCcsHeader("$vsdir/tym.hdr", $itype,$vtype,[$NT,$ymax+1])
     or $vs->logconfess("$logas: failed to save CCS header $vsdir/tym.hdr: $!");
 
-  ##-- create: tym: cleanup
+  ##-- tym: cleanup
   undef $c2date;
   undef $c2d;
   undef $d2c;
 
-  ##-- create: tdm: pointers
+  ##-- create: aux: tdm: pointers
   $vs->vlog($logCreate, "$logas: creating tdm matrix Harwell-Boeing pointers");
   my ($ptr0) = $tdm->getptr(0);
   $ptr0      = $ptr0->convert($itype) if ($ptr0->type != $itype);
@@ -810,7 +831,7 @@ sub union {
   my $NA       = scalar @{$vs->{attrs}};
   my $tdm0file = "$vsdir/utdm0.dat";
   $vs->vlog($logCreate, "union(): extracting attribute-document data to $tdm0file");
-  my $tdm0fh   = opencmd("|-:raw", 'sort', (map {"-nk$_"} (1..($NA+1))), "-o", $tdm0file)
+  my $tdm0fh   = opencmd("|-:raw", 'sort', (map {"-nk$_"} (1..($NA+1))), sortJobs(), "-o", $tdm0file)
     or $vs->logconfess("union(): failed to create pipe to sort for tempfile $tdm0file: $!");
   my $Doff = 0;
   foreach my $dbai (grep {$dbargs->[$_]{tdf}} (0..$#$dbargs)) {
